
#include "BasicSolver.hpp"
#include <algorithm>
#include "Logging.hpp"
#include "AssignInfos.hpp"

const char*  basicsolver_cpp_version(){return "$Revision: 1.49 $";}


BasicSolver::BasicSolver(TGlobals& globals):
	glob						(globals),

	var_act_comp				(),
    var_act_hp              	( glob.sets.var_max_act , glob.sets.var_decay, var_act_comp),
    sel_strt					(FIX_VAR),
    cls_mode 					( true),
    ass_info					(vals,var_act_hp,glob.sets.bump_agil),
    watching					(),
    queue_begin                 (0),
    //real_units_end				(0),

    simplify_assign_idx         (0),
    simpl_next					(0),
    max_lbd_gc					(2),
    set_bcp_shrts				(true),
    lbd_total					(0),
    lbd_confl					(0)
{
	reserveForVariables(0,true);
	setUnit(TRUE_LIT,CORE_DONT_CARE,true);
	asrt(vals.litTrue(TRUE_LIT));
}


int32_t	BasicSolver::bcp(){
	T_ON(prop);

    int32_t        	nbr_pivot_lits	= BCP_SUCCESS; //indicates no conflict by now
    uint32_t        num_props		= 0;

    while(queue_begin < assign_stck.size()){
        ++num_props;
        nbr_pivot_lits = bcpPropagateLit(assign_stck[queue_begin++].flipped());
    }

    cnt_stat(props,num_props);
    LOG2("Did " << num_props << " props");

    T_OFF(prop);
    return nbr_pivot_lits;
}





int32_t	BasicSolver::bcpPropagateLit(const Literal& false_lit){
	 LOG2("Propagate lit " << false_lit.flipped().toDIMACS());
	WIter<WtchEle> wi	= watching.getWtchOfW(false_lit);//watched clauses that have to be checked when lit gets false
	WtchEle wt; uint32_t eq_lbd;

	while(wi.getNext(wt)){

		//might be any other literal of this clause except the watched literal
        Literal cach_lit = wt.getLit();   asrt(cach_lit != false_lit);

        if(vals.litTrue(cach_lit)){wi.keep(wt); continue;}

        ClsP clp = wt.getClsP();

        if(wt.isBcpShrt()){
        	wi		.keep(wt);
        	eq_lbd	= Reason::LBD_MAX;

        	asrt(cach_lit == glob.getCls(clp).otherWatcher(false_lit));


		}else{

	        cach_lit		= glob.getCls(clp).otherWatcher(false_lit);

			//it might be that the shortcut literal in 'wt' does not point to the actual watched literal anymore
			if(vals.litTrue(cach_lit)){
				wi.keep(WtchEle(wt,cach_lit));
			    continue;
			}

			Cls& cl  = glob.getCls(clp);
			if(bcpFindNewWatcher(false_lit,cach_lit,wt)){continue;}

			LOG2("No new watcher found for lit "<<false_lit.toDIMACS() << " other is " <<cach_lit.toDIMACS());
			wi		.keep(WtchEle(wt,cach_lit));
			eq_lbd 	= wt.isLrnt() ? getLBD(cl,wt) : Reason::LBD_MAX;
		}


        if (vals.litFalse(cach_lit)){
	    	return bcpFinishOnConflict(false_lit,wi,clp);}

        LOG2("Enqueue Literal " << cach_lit.toDIMACS());
	    enqueue(cach_lit, clp, eq_lbd);
	    asrt(!wt.isLrnt() || eq_lbd == ass_info.reasonLBD(cach_lit.toVar()));
	}
	LOG2("Finish propagation of " << false_lit.flipped().toDIMACS());
	watching.finish(false_lit,wi);
    return BCP_SUCCESS;
}


int32_t BasicSolver::bcpFinishOnConflict(const Literal& false_lit,  WIter<WtchEle>& iter, const ClsP& clp){
	LOG2("Finish on conflict with lit "<<false_lit.toDIMACS()); asrt(var_flags.allCleared());

	Cls& cl  		= glob.getCls(clp);
	queue_begin 	= assign_stck.size(); // to finish outer while loop

	//copy rest
	WtchEle cp;
	while(iter.getNext(cp)){iter.keep(cp);}
	watching.finish(false_lit,iter);

	//init conflict lemma stuff
	lemma					.resize(1,INVALID_LIT);
	lemma_zro				.clear();
	confl_lbd 				= 1;
	conflVars				.clear();

	const uint32_t dlev		= decisionLevel();
	asrt(vals.litFalse(false_lit)); asrt(vals.litFalse(cl.otherWatcher(false_lit)));

	//mark core
	if(dlev == 0){ asrt(asrtAllFalse(false_lit,clp,0));
		//cout << " do initial mark of cls " << clp.idx() << endl;
		FOR_ALL_LITS_DO(false_lit,cl,setVarOf);
		//cout << " fini initial mark of cls " << clp.idx() << endl;
		return 0;
	}

	//mark lits for first UIP
	int nbr_mrkd = firstUIPMrkLit(false_lit,dlev) + firstUIPMrkLit(cl.otherWatcher(false_lit),dlev);

	FOR_UW_LITS_STRT(cl,lit);
		nbr_mrkd += firstUIPMrkLit(lit,dlev);
	FOR_UW_LITS_END; asrt(nbr_mrkd != 1); asrt(nbr_mrkd != 0);
	LOG2("Finish on conflict marked "<<nbr_mrkd << " lits at level " << dlev);

	return nbr_mrkd;

}




void BasicSolver::printModel(std::ostream& out){
	out << endl << "v "; Variable max (ass_info.varsCount());
	for(Variable v(1);v<max;v=Variable(v.idx()+1)){	asrt(vals.varKnwn(v));
		if(v >= sel_strt){break;}
		out << v.toLit(vals.varPrevValWasTrue(v)).toDIMACS() << " " ;
	}
	out << " 0" << endl;
}

void BasicSolver::printCore(vector<uint32_t>* cls_ids){
	asrt(min_core.size() + core_cands.size() != 0);

	const bool cons = cls_ids == NULL;
	if(core_cands.empty()){
		if(cons){ cout << "c Core is minimal" << endl;}
	}else{
		for(uint32_t i=0;i!=core_cands.size();++i){
			min_core.push_back(core_cands[i].lit);
		}
		core_cands.clear();
		if(cons){ cout << "c Core not minimised" << endl;}
	}

	if(cons){ cout << endl << "v ";}

	uint32_t csz = 0;
	for(uint32_t i= min_core.size();i--;){
		Literal lt		= min_core[i];
		Variable v 		= lt.toVar();
		Reason r 		= ass_info.varAssignReason(v); asrt(r.isUnit());
		uint32_t cid	= r.getUnitClsID();
		if(cid == CORE_DONT_CARE) continue;
		asrt(cid != 0); asrt(cid != CORE_TEST_CAND);
		++csz;
		if(cons){cout <<	cid << " ";}
		else{cls_ids->push_back(cid);}
	}

	if(cons){
		cout << " 0" << endl;
		cout << endl << "c Core has size " << csz << endl;
	}
}




bool 		BasicSolver::parseGroups(IFStream& in,const uint32_t vars, const uint32_t expect){
	reserveForVariables(vars,true);
	reserveForVariables(vars + expect,false);
	cls_mode = false;

	vector<Literal> plits; int32_t  prsd_l; int32_t grp_id;
	Literal lit = FALSE_LIT;

	while(in.hasNext()){
		if(in.currEquals('c')) {in.skipLine(); continue;}

		in.skipWhitespace();
		if(in.currEquals('{')){
			++in;
			grp_id = in.getNextInt();
			in.skipWhitespace();
		}else{grp_id = -1;}

		if(grp_id < 0 || !in.currEquals('}')){
			cout << "c PARSE ERROR in file " << glob.sets.cnf_file_name << endl,
			exit(EXIT_FAILURE);
		}
		++in;

		//parse new clause
		plits.clear();
		while((prsd_l = in.getNextInt()) != 0){
			lit = Literal::DIMACS2Literal(prsd_l); asrt(prsd_l == lit.toDIMACS());
			plits.push_back(lit);
		}

		//check for empty clause
		if(plits.empty()){
			if(grp_id != 0){//add dummy variable to print proper group id
				Variable dummy(vars + expect + 2);
				reserveForVariables(dummy.idx(),false);
				setUnit(dummy.toNegLit(),grp_id,true);
				min_core	.push_back(dummy.toNegLit());}
			return glob	.setInvalid("Added empty clause at initialisation");
		}

		uint32_t ukwn = validateCls(plits);
		if(ukwn == max32){continue;}


		if(grp_id != (int32_t)CORE_DONT_CARE){
			Variable grp = Variable(vars+grp_id);
			reserveForVariables(grp.idx(),false);
			if(!vals.varKnwn(grp)){
				LOG2("Init group selector variable " << grp.toDIMACS());
				setUnit(grp.toNegLit(),grp_id,false);
				var_act_hp.removeIfContained(grp);
			}
			plits.push_back(grp.toPosLit());
		}

		if(ukwn == 0){
	    	for(uint32_t i=0;i!=plits.size();++i){setVarOf(plits[i]);}

			computeCore();

			return glob.setInvalid("Added clause was empty due to unit assignments");
		}

		asrt(vals.litUkwn(plits[0]));

		if(plits.size() == 1){ asrt(grp_id == 0);
			setUnit(plits[0],CORE_DONT_CARE,true); //take unit in min core since group 0 does not matter
		}else{
			if(plits.size() == 2){plits.push_back(FALSE_LIT);}
		   	clsCreate(plits,ukwn,false,0,true);
		}
	}

	if(bcp() != BCP_SUCCESS){
		computeCore();
		return glob.setInvalid("Empty Clause when propagating initial unit clauses");
	}

	return true;

}

BasicSolver::~BasicSolver(){}



bool BasicSolver::parseRegular(IFStream& in,const uint32_t vars, const uint32_t clses) {
	asrt(min_core.empty()); asrt(core_cands.empty());
	reserveForVariables(vars, true);
	reserveForVariables(vars + clses / 2, false);
	cls_mode = true;

	vector<Literal> parsed_lits; int32_t  parsed_lit;
	Variable next_free 	= sel_strt;
	uint32_t cls_id 	= 0;
	Literal lit 		= FALSE_LIT; //uint32_t c = 0; uint64_t l = 0; uint32_t lts = 0;

	while(in.hasNext()){
		if(in.currEquals('c')) {in.skipLine(); continue;}

		while((parsed_lit = in.getNextInt()) != 0){
			lit = Literal::DIMACS2Literal(parsed_lit); asrt(parsed_lit == lit.toDIMACS());
			parsed_lits.push_back(lit);
		}

		++cls_id;

		if(haveToAllocVars(next_free.idx())){
			double new_vars = next_free.idx() + 1 - sel_strt.idx(); asrt(new_vars<=cls_id);
			double expct	= (new_vars / (double)cls_id) * clses; asrt(expct>=0);asrt(expct<=clses);
			reserveForVariables(vars + 1 + (uint32_t)expct,false);
		}

		if(!addPrsdClause(parsed_lits,cls_id,next_free)){
			LOG1("Formula contains an empty clause!! - is unsatisfiable!\n");
			return glob.setInvalid("Empty Clause in parsed formula");
		}
		parsed_lits.clear();
	}

	if(bcp() != BCP_SUCCESS){
		doSimplify(false,false);
		computeCore();
		return glob.setInvalid("Empty Clause when propagating initial unit clauses");
	}

	doSimplify(true,false); //to make sure that clauses with true (unit) lits are removed since core computation may flip this lit

	return true;
}



bool BasicSolver::parseDIMACS(IFStream& in) {
	uint32_t grp = 0;

	while(in.hasNext()){

		if(in.currEquals('c')) {in.skipLine();continue;}

		if(!in.currEquals('p')){
			LOG1("Did not find problem line");
			exit(EXIT_FAILURE);
		}

		++in; in.skipWhitespace();

		if(in.currEquals('g')){
			if (in == "gcnf"){ grp = 1;  break;}

		}else if(in == "cnf"){break;}

		cout << "c PARSE ERROR in file " << glob.sets.cnf_file_name << endl,
		exit(EXIT_FAILURE);
	}

	uint32_t vrs 		= in.getNextInt();
	uint32_t cls 		= in.getNextInt();
	reserveForClauses	(cls);
	sel_strt			= Variable(vrs+1);

	if(grp){
		grp = in.getNextInt();
		in.skipLine();
		return parseGroups(in,vrs,grp);
	}

	in.skipLine();
	return parseRegular(in,vrs,cls);
}

struct ClsSrt{
	const VarVal& vals;
	ClsSrt(const VarVal& v):vals(v){};
	bool operator()(const Literal& a, const Literal& b) const{
		asrt(a.toVar().idx() < vals.count()); asrt(b.toVar().idx() < vals.count());
		return vals.litFalse(a) == vals.litFalse(b) ? a < b : vals.litFalse(b);
	}
};


uint32_t	BasicSolver::validateCls(vector<Literal>& lits){
	asrt(decisionLevel() == 0); asrt(glob.isValid()); asrt(core_cands.empty());asrt(min_core.empty()); asrt(!lits.empty());

	// Check if clause is satisfied and remove false/duplicate literals:
	std::sort(lits.begin(),lits.end(),ClsSrt(vals));

	Literal last(INVALID_LIT); uint32_t sz = 0;
	uint32_t ukwn = 0;
	for(uint32_t i=0;i != lits.size();++i){
		asrt(lits[i].toVar() < sel_strt);

		if(vals.litTrue(lits[i]) || lits[i] == last.flipped()){  return max32;}

		if(lits[i] != last){
		   lits[sz++] = last = lits[i];
		   ukwn += vals.litUkwn(last);
		}
	}
	lits.resize(sz);
	return ukwn;
}


bool BasicSolver::addPrsdClause(vector<Literal>& lits, const uint32_t cls_id, Variable& sel_var){

    asrt(decisionLevel() == 0); asrt(glob.isValid()); asrt(core_cands.empty());asrt(min_core.empty()); asrt(sel_var.idx() < vals.count());

    if (lits.empty()){
    	setUnit(sel_var.toNegLit(),cls_id,true);
    	min_core	.push_back(sel_var.toNegLit());
    	return glob	.setInvalid("Added empty clause");
    }

    uint32_t ukwn = validateCls(lits);
    if(ukwn == max32){return true;}

    if(ukwn == 0){ asrt(var_flags.allCleared());

    	doSimplify(false,false); //to make sure that clauses with true lits are removed since core computation may flip this lit

    	setUnit(sel_var.toNegLit(),cls_id,true);

    	for(uint32_t i=0;i!=lits.size();++i){setVarOf(lits[i]);}

    	computeCore();

    	return glob.setInvalid("Added clause was empty due to unit assignments");
    }

    asrt(vals.litUkwn(lits[0]));

    if(lits.size() == 1){
    	setUnit(lits[0],cls_id,false);
    	return true;
    }

    //dont propagate selector variable immediately to enable smaller proofs
    setUnit(sel_var.toNegLit(),cls_id,false);
    var_act_hp.removeIfContained(sel_var);

   	lits	.push_back(sel_var.toPosLit());
   	clsCreate(lits,ukwn,false,0,true);


    sel_var 	= Variable(sel_var.idx() + 1); // for next clause

    return true;
}




uint32_t	BasicSolver::getLBD(const Cls& cls, const WtchEle& we){ asrt(we.isLrnt());
	return (we.getLBDHigh() << 2) + cls.getLbdLow() + 1; //what a hack: part of the lbd value is stored in the watchers the other part within the clause
}

ClsP BasicSolver::clsCreate(vector<Literal>& lits, const uint32_t ufix, const bool learnt,uint32_t lbd,bool set_shrt){

	const uint32_t sz = lits.size(); asrt(ufix <= sz);
    asrt(sz > 2); asrt(learnt || lbd==0); asrt(lbd <= Reason::LBD_MAX); asrt(!learnt || lbd > 0); asrt(!vals.litFalse(lits[0]) || !vals.litFalse(lits[1]));

    ClsP loc			= glob.c_info.alloc(Cls::allocSizeFor(sz));

	if(loc.isNull()){
		cout << endl << "c !!!OUT OF MEMORY!!!" << endl;
		exit(EXIT_FAILURE);
	}

	lbd 				-= learnt; asrt(lbd < Reason::LBD_MAX);
	uint32_t lbd_high 	= lbd >> 2;
	uint32_t lbd_low	= lbd & 3;

    new (glob.c_info.getRealPtr<Cls>(loc)) 	Cls(lits[0],lits[1],lbd_low);
    Cls& cls	= glob.getCls(loc);
    ClsLit* it	= cls.uwBeg();

    for(uint32_t i = 2;i != sz;++i) {*it++ = ClsLit(lits[i],true);} asrt(it != cls.uwBeg());
    (it-1)->setEnd();

    if(ufix>2 || !set_shrt){
    	if(set_shrt){cls	.setPropEnd(ufix);}
    	watching			.clsAddWtched(lits[0],lits[1],loc,learnt,lbd_high);
    	watching			.clsAddWtched(lits[1],lits[0],loc,learnt,lbd_high);
    	cnt_stat(lcls ,1);

    	//aux_var(WtchEle we1 = watching.getLastOf(lits[0]);) asrt(we1.isLrnt() == learnt);
    	//aux_var(WtchEle we2 = watching.getLastOf(lits[1]);) asrt(we2.isLrnt() == learnt);
    	//asrt(!learnt || (getLBD(cls,we1) == lbd + 1));
    	//asrt(!learnt || (getLBD(cls,we2) == lbd + 1));

    }else{
    	cls					.clearLBDPrt();
    	watching			.clsAddBinPrt(lits[0],lits[1],loc);
    	watching			.clsAddBinPrt(lits[1],lits[0],loc);
    	cnt_stat(bin_cls ,1);
    }
    cnt_stat(lts,sz);

    return loc;
}


/*
bool BasicSolver::extendMinCore(const Literal& unit){
	LOG1("Set unit for Core" << unit.toDIMACS());
	if(ass_info.varIsUnit(unit.toVar())){
		takeCandInCore(unit.toVar());
		core_cands.clear(); LOG1("Take flipped unit in min_core and terminate");
		return false;
	}
	LOG1("Take unit in min_core but as dont care");
	setUnit(unit,CORE_DONT_CARE,true); //reason for this unit is already marked to be in minimal core
	return true;
}*/

bool BasicSolver::reviveClauses(){

	LOG2("Revive clauses");
	Literal fail = INVALID_LIT; uint32_t rlbd;

	while(glob.cemetery.getNext(lemma,rlbd,vals)){
		if(lemma.empty()){	core_cands.clear(); return false;}

		const bool lrnt 	= rlbd != 0;
		const uint32_t rsz 	= lemma.size();
		if(rsz == 1){ LOG1("Revival of clauses caused unit " << lemma[0].toDIMACS());
			Literal unit = lemma[0];
			if(ass_info.varIsUnit(unit.toVar())){
				LOG1("Conflict with core candidate in clause revival " << unit.toDIMACS());
				fail = unit;
			}else{
				LOG1("Enqueue Literal in clause revival " << unit.toDIMACS());
				setUnit(unit,CORE_DONT_CARE,true);
			}
		}else{
			if(rsz == 2){lemma.push_back(FALSE_LIT);}
			clsCreate(lemma,rsz,lrnt,rlbd,true);
			if(rlbd != 0){glob.registrLBD(rlbd);}
			LOG1("Revival of clause with size " << lemma.size() << " first " << lemma[0].toDIMACS() << " second " << lemma[1].toDIMACS());
		}
	}

	//test last candidate
	if(fail != INVALID_LIT){
		keepThisCandOnly(fail.flipped());
		return false;}

	return true;
}




bool BasicSolver::propagateMinCore(){//propagate while cand lits are not in assign_stck
	LOG2("propMinCore start with " << queue_begin); asrt(asrtUnitsEnquedOrInCoreCands());
	WtchEle wt; asrt(decisionLevel()==0);
	uint32_t        num_props		= 0;
	Literal fail = INVALID_LIT;


	while(queue_begin < assign_stck.size()){
		++num_props;
		Literal false_lit 	= assign_stck[queue_begin++].flipped();
		WIter<WtchEle> wi	= watching.getWtchOfW(false_lit);//watched clauses that have to be checked when lit gets false
		LOG2("In propMinCore prop "<<false_lit.toDIMACS() << " at index " << queue_begin -1);

		while(wi.getNext(wt)){

			ClsP clp	= wt.getClsP();
			Cls& cl  	= glob.getCls(clp);
			Literal ow 	= cl.otherWatcher(false_lit);

			if(vals.litTrue(ow)){wi.keep(WtchEle(wt,ow)); continue;}

			FOR_UW_LITS_STRT(cl,cnd);
				if(!vals.litFalse(cnd)){
					cl			.changeWatchers(ow,clit,false_lit); LOG2("Found new watcher "<<cl.otherWatcher(ow).toDIMACS());
					watching	.changeWatching(wt,ow,cl.otherWatcher(ow));
					goto NEXT_CLS;}
			FOR_UW_LITS_END;


			LOG2("No new watcher in propMinCore "<<false_lit.toDIMACS() << " other is " <<ow.toDIMACS());
			wi.keep(WtchEle(wt,ow));

			if(vals.litUkwn(ow)){
				if(ass_info.varIsUnit(ow.toVar())){
					LOG1("Conflict with core candidate in min core propagation " << ow.toDIMACS());
					fail = ow;
				}else{
					LOG1("Enqueue Literal in min core propagation " << ow.toDIMACS());
					setUnit(ow,CORE_DONT_CARE,true);
				}
			}else{
				LOG0("Current min_core is unsatisfiable. Tell it to the world and leave!");
				while(wi.getNext(wt)){wi.keep(wt);}
				watching	.finish(false_lit,wi);
				core_cands	.clear(); //min_core is minimal now
				return false;
			}

			NEXT_CLS:;
		}

		LOG2("Finish propagation of " << false_lit.flipped().toDIMACS());
		watching.finish(false_lit,wi);
	}

	if(fail != INVALID_LIT){
		keepThisCandOnly(fail.flipped());
		return false;
		//test last candidate
	}

	cnt_stat(props,num_props); LOG2("Did " << num_props << " props of min core now " << queue_begin);
	return true;
}

//todo heuristic on order
struct CandSort{
	const var_hp_t& hp;
	const AssReason& ai;
	CandSort(const var_hp_t& h,const AssReason& ar):hp(h),ai(ar){}
	bool operator()(const CoreCand& a, const CoreCand& b) const{
		Variable va = a.lit.toVar();  Variable vb = b.lit.toVar();

		return ai.didQuickTest(va) == ai.didQuickTest(vb)
				?	hp.getValue(va) > hp.getValue(vb)
				: 	ai.didQuickTest(va);}

		/*return ai.didQuickTest(va) ? (!ai.didQuickTest(vb) ||  hp.getValue(va) < hp.getValue(vb))
									: (ai.didQuickTest(vb) &&  hp.getValue(va) > hp.getValue(vb)) ;}*/


};




void		BasicSolver::minimiseCore(){asrt(!glob.isValid());

	CandSort act_comp(var_act_hp,ass_info);
	vector<Literal> revive; asrt(asrtUnitsEnquedOrInCoreCands()); asrt(queue_begin <=1);

	bool quick;
	int32_t next_compl_check = 0; //simplify immediately

	//glob.sets.restart_type 		= glob.sets.RSTRT_LUBY;
	//glob.sets.lbd_rest_buffr	= 1024;

	//Assym assy(glob,var_flags,watching,ass_info);

	while(!core_cands.empty()){
		asrt(asrtUnitsEnquedOrInCoreCands());

		prepareNewRound(4* core_cands.size());

		//at this point only literals are enqueued that are either known to be in the minimal core, or are known to be not in the minimal core

		//revive deleted clauses
		if(!reviveClauses()) continue;

		//simplify instance if new core was found or after some extensions to the core
		if(next_compl_check <= 0){
			//asrt(!glob.isValid()); asrt(asrtUnitsEnquedOrInCoreCands());
			LOG1("Core with size " << min_core.size() << " sure " << " and " << core_cands.size() << " candidates");//. Given reason: " << glob.whyInvalid());

			std::sort(core_cands.begin(),core_cands.end(),act_comp);
			next_compl_check = (core_cands.size() / 20) + 100;

			doSimplify(true,false);

		}else{LOG2("New round without conflict");}

		//Get test one core cand
		CoreCand cnd	= core_cands.back(); asrt(vals.litUkwn(cnd.lit));
		LOG1("Choose one cand for core test " << cnd.lit.toDIMACS());

		if(core_cands.size()<2){ //current test candidate is the last SAT check
			quick = false;		//dont propagate min core -> this might cause an endless loop with "keepThisCandOnly"

		}else{
			//propagate -> if conflict then current min_core is minimal or one element away (test last one)
			if(!propagateMinCore()){continue;}

			quick = ass_info.setCoreTestNew(cnd.lit);
		}
		asrt(core_cands.back().lit == cnd.lit);
		core_cands		.pop_back();


		//simplify clauses with literals that are known to be in the core anyway
		//todo

		LOG2("Enqueue core cands");
		const uint32_t ass_fix = assign_stck.size(); LOG2("Fix Units are at now "<< ass_fix);

		//enqueue core cands
		for(uint32_t t=0; t < core_cands.size(); ++t){ //for all but last (with least activity)
			CoreCand& cc = core_cands[t]; asrt(vals.litUkwn(cc.lit));
			setUnit(cc.lit,cc.cls_id,false);
		}

		//enqueue test candidate
		LOG1("Check sat with flipped " << cnd.lit.toDIMACS() << " quick is " << quick);
		setUnit(cnd.lit.flipped(),CORE_TEST_CAND,false); asrt(asrtNotContained(cnd.lit));

		asrt(queue_begin <= ass_fix);

		TrBool res = isSATInnr( quick ? 2 : max32);
		LOG2("Got result of sat check:" << res.toStrng());

		//if false, new core was computed -> we start again
		if(res.isFalse()){next_compl_check = 0; LOG1("Found smaller core"); continue;}

		LOG2("Clear level 0 assigns "); asrt(res.isUnknown() || asrtAllClsesSAT());
		clearAssignsHard(ass_fix); asrt(queue_begin <= ass_fix);


		if(res.isTrue()){
			LOG1("Increase minimal core by lit " << cnd.lit.toDIMACS() << " with id " << cnd.cls_id);
			min_core	.push_back(cnd.lit);
			setUnit(cnd.lit,cnd.cls_id,true);

			--next_compl_check;

		}else{//timeout
			LOG1("Quick test was too slow of " << cnd.lit.toDIMACS() << " with id " << cnd.cls_id);

			core_cands.push_front(cnd);
			if(glob.stats.cpuUsrTime() > glob.sets.timeout) {return;}
		}

	}
}


void BasicSolver::keepThisCandOnly(const Literal& lit){
	Variable var = lit.toVar();
	LOG1("There is only one cand " << lit.toDIMACS() << " left");

	asrt(!core_cands.empty()); asrt(!vals.varKnwn(var)); asrt(!ass_info.varUnitFix(var));
	asrt(asrtVarInCoreCands(var)); asrt(vals.varPrevValWasTrue(var) == lit.isPos()); asrt(asrtUnitsEnquedOrInCoreCands());

	uint32_t cls_id = CORE_DONT_CARE;
	for(uint32_t i=0;i!=core_cands.size();++i){
		Literal cnd = core_cands[i].lit; asrt(cnd != lit.flipped());

		if(cnd == lit){						asrt(cls_id == CORE_DONT_CARE);
			cls_id = core_cands[i].cls_id; 	asrt(cls_id != CORE_DONT_CARE);

		}else{
			asrt(ass_info.varIsUnit(cnd.toVar())); asrt(!ass_info.varUnitFix(cnd.toVar()));
			asrt(core_cands[i].cls_id != CORE_TEST_CAND); asrt(asrtNotContained(cnd));

			LOG2("Set lit " << lit.toDIMACS() << " as never to be in min core");
			setUnit(cnd.flipped(),CORE_DONT_CARE,true);
		}
	}
	asrt(cls_id != CORE_DONT_CARE);

	core_cands.clear();
	core_cands.push_back(CoreCand(lit,cls_id));
	queue_begin = 1; //cause propagation from scratch
}


void BasicSolver::computeCore(){
	asrt(decisionLevel() == 0); asrt(!assign_stck.empty()); asrt(assign_stck[0]==TRUE_LIT);

	min_core.clear(); core_cands.clear();

	//analyse backwards
	for(uint32_t i=assign_stck.size();--i;){
		Literal lit = assign_stck[i];
		if(!var_flags.isSet(lit.toVar())) continue;

		const Reason& r = ass_info.varAssignReason(lit.toVar()); asrt(!r.isDecision());

		//cout << " handle lit " << lit.toDIMACS() << endl;

		if(r.isClsReason()){
			ClsP clp 	= r.getClsReason();
			Cls& cls	= glob.getCls(clp);
			LOG2("analyse cls back at 0 reason of " << lit.toDIMACS());
			//cout << " do mark of cls " << clp.idx() << endl;

			LOG2("analyse cls back at 0 watcher is " << cls.otherWatcher(lit).toDIMACS());
			setVarOf(cls.otherWatcher(lit)); asrt(vals.litFalse(cls.otherWatcher(lit)));
			FOR_UW_LITS_STRT(cls,li);
				LOG2("analyse cls back at 0 uw is " << li.toDIMACS());
				setVarOf(li); asrt(vals.litFalse(li));
			FOR_UW_LITS_END;

			LOG2("analyse cls back at 0 fini of " << lit.toDIMACS());
			//cout << " fini mark of cls " << clp.idx() << endl;
		}
	}
	var_flags.clear(FALSE_LIT.toVar());

	//keep core assignments. split  vars that are in the min core for sure, and candidates for min core
	uint32_t cp = queue_begin = 1;
	for(uint32_t i=1;i<assign_stck.size();++i){
		Literal lit 	= assign_stck[i];
		Variable var	= lit.toVar();
		const Reason& r = ass_info.varAssignReason(var); asrt(!r.isDecision());

		if(r.isUnit()){
			uint32_t cid= r.getUnitClsID();

			if(r.isUnitFix()){ 				asrt(cid != CORE_TEST_CAND);
				LOG2("found lit " << lit.toDIMACS() << " with state known");
				if(cid != CORE_DONT_CARE){ //for group mode
					LOG2(" found lit " << lit.toDIMACS() << " in min core");
					min_core		.push_back(lit);
				}

			}else if(var_flags.isSet(var)){		asrt(cid != CORE_TEST_CAND); asrt(cid != CORE_DONT_CARE);

				LOG2("found lit " << lit.toDIMACS() << " to be core candidate");
				ass_info			.clearL0Value(var);
				core_cands			.push_back(CoreCand(lit,cid));
				goto				DONT_KEEP_ENQUEUED; //to allow minimisation of core candidates

			}else{
				LOG2("found lit " << lit.toDIMACS() << " who will never be in min core");
				if(cid != CORE_TEST_CAND){asrt(asrtNotContained(lit));
					lit = lit.flipped();} //test cand is already flipped
				ass_info			.setValIniUnit(lit,CORE_DONT_CARE,true);
			}
			assign_stck[cp++] 	= lit;

		}else{ asrt(r.isClsReason());
			LOG2("found lit " << lit.toDIMACS() << " but dont keep at 0");
			ass_info			.clearL0Value(var);
			var_act_hp			.restoreElementIf(var);
		}

		DONT_KEEP_ENQUEUED:;
		var_flags	.clear(var);
	}
	assign_stck.resize(cp);

	glob.stats.cnt_core_lts = min_core.size() + core_cands.size();
	LOG1(" Computed core with at least " << min_core.size() << " and cands " << core_cands.size());
	asrt(var_flags.allCleared());
}





TrBool BasicSolver::isSATInnr(const uint32_t max_strts){
	asrt(vals.litTrue(TRUE_LIT));asrt(glob.isValid()); asrt(decisionLevel()==0); asrt(var_flags.allCleared());

	lbd_total	= 0;
	lbd_confl	= 0;
	lbd_history .clear();

	 if(glob.sets.restart_type == glob.sets.RSTRT_LUBY){
		luby_restarts	= LubyRestarts(glob.sets.lbd_rest_buffr);
		lbd_history		.resize(1);}
	else{
		lbd_history		.resize(glob.sets.lbd_rest_buffr);
	}


	 //TrBool status = glob.sets.do_simpl  ? preProcess() : UNKNOWN_TrBool;
	TrBool status = UNKNOWN_TrBool;

	for(uint32_t s=0; s != max_strts && (glob.stats.cpuUsrTime() <= glob.sets.timeout); ++s){asrt(glob.isValid());
		uint32_t cfl_mx 		= luby_restarts.nextRestartGetConflictCount();
		status 					= search(cfl_mx);

		if(glob.sets.print_console){	glob.stats.printProgress();}

		if(status.isKnown()) 	break;

		backjumpKeepLevel(0);
	}

	return status;
}




TrBool BasicSolver::isSAT(){
	asrt(vals.litTrue(TRUE_LIT));
	if (!glob.isValid())          return FALSE_TrBool;

	//next_ext_var			= Variable(ass_info.varsCount());
    //var_act_comp.ext_border= next_ext_var; LOG1("First extended is " << next_ext_var.toDIMACS());


//glucose
    uint32_t max_l  = glob.stats.cnt_lcls / 3;
    if(max_l  <glob.sets.max_learnts) {
    	glob.sets.max_learnts = max_l >> 1;
    	if(glob.sets.max_learnts < 5000) glob.sets.max_learnts = 5000;
    }
    LOG1("Max learnts limit is "<< glob.sets.max_learnts << " incremented by " << glob.sets.max_learnts_inc);



    return isSATInnr(max32);
}


void    BasicSolver::backjumpUndoAssigns (uint32_t assign_idx){

	Variable x = INVALID_VAR;
	asrt(assign_stck.size() >= assign_idx);

	queue_begin 	= assign_stck.size();

	while(queue_begin != assign_idx ){
		x  = assign_stck[--queue_begin].toVar();
		vals.varClearValue(x);
		var_act_hp.restoreElementIf(x);}

	assign_stck.resize(queue_begin);
}

void	BasicSolver::clearAssignsHard(uint32_t from){

	backjumpKeepLevel(0);

	if(from >= assign_stck.size()) return;

	uint32_t cp  = queue_begin = from;
	for(; from < assign_stck.size();++from){
		Literal l		= assign_stck[from];
		const Reason& r	= ass_info.varAssignReason(l.toVar());

		if(r.isUnitFix()){ asrt(ass_info.varUnitClsID(l.toVar())==CORE_DONT_CARE);
			LOG2("Keep new unit clause in hard clear " << assign_stck[from].toDIMACS());
			assign_stck[cp++] = l;
			continue;
		}

		LOG2("Hard clear of " << assign_stck[from].toDIMACS());
		ass_info		.clearL0Value(l.toVar());
		if(!r.isUnit()){
			var_act_hp	.restoreElementIf(l.toVar());}
	}

	assign_stck.resize(cp);
	LOG2("In clearAssignsHard reset stck to "<<cp);

}




TrBool BasicSolver::search(uint32_t max_confl ){

    asrt(glob.isValid());
    TrBool   result	= UNKNOWN_TrBool;
    const uint32_t abs_confl = lbd_confl + max_confl;
    bool restart;

    cnt_stat(strts,1);

    while(true ){

        int32_t bcp_res    =     bcp();

        if(bcp_res == BCP_SUCCESS){         //there is no conflict

        	restart = (glob.sets.restart_type == glob.sets.RSTRT_LUBY) ?
        			lbd_confl >= abs_confl : //luby
					lbd_history.enoughData()&& ((lbd_history.getAvg()*0.7) > (lbd_total / (lbd_confl+1))); //glucose

        	if(restart) {
        		LOG2("Clear Lbd History");
        		lbd_history.clear();
				if (ass_info.lessAgilThan(2500000)){ break;} //Biere
				opt_istat(rstrts_skip,1);
        	}


        	if(shallGC()){learntsGC();}

             // Simplify the set of problem clauses
            if (decisionLevel() == 0 && shallSimplify()){doSimplify(false,true);}

            // New decision:
            Literal next    = nextDecision();

            //check if model is found
            if (next == INVALID_LIT)    {return TRUE_TrBool;}

            // Increase decision level and enqueue 'next'
            asrt(queue_begin == assign_stck.size()); asrt(vals.litUkwn(next));
            LOG2("Next decision lit " << next.toDIMACS());

            decide(next);


        }else{                         //there is a conflict
        	asrt(bcp_res != 1);

        	if(bcp_res == 0){
        		computeCore();
        		return TrBool(glob.setInvalid("Conflict at level 0"));
        	}

        	asrt(decisionLevel() != 0);
			firstUIP(bcp_res);

			asrt(ass_info.litAssLev(lemma[0]) == decisionLevel());
			asrt(!var_flags.isSet(lemma[0].toVar())); asrt(!isSetLevel(decisionLevel())); asrt(!isSetLevel(0));

			//glucose
			lbd_history	.keep(confl_lbd);
			lbd_total	+= confl_lbd;
			++lbd_confl;

			if(confl_lbd >= Reason::LBD_MAX){
				confl_lbd=Reason::LBD_MAX;

			}else if(glob.sets.bump_glucose){
				for(uint32_t i=conflVars.size();i--;){
					const Reason& ri = ass_info.varAssignReason(conflVars[i]); asrt(vals.varKnwn(conflVars[i])); asrt(!ass_info.varIsUnit(conflVars[i]));
					if(ri.getLBD() < confl_lbd){ varBump(conflVars[i]);}
				}
			}

			reduceSelfImpl();

			uint32_t ufix = lemma.size();
			resolveZeros();

			uint32_t btrk_lev = backtrackLevel(); asrt(btrk_lev < decisionLevel());

			opt_istat(sum_bck_jmp,decisionLevel() - btrk_lev);
			opt_istat(sum_confl_depth,decisionLevel());

			backjumpKeepLevel(btrk_lev);	asrt(vals.litUkwn(lemma[0]));
			var_act_hp.decayAll();

			if(lemma.size() == 1){ LOG1("Learnt unit clause " << lemma[0].toDIMACS() << " Take in core as dont care");
				setUnit(lemma[0],CORE_DONT_CARE,true);
				continue;}

			bool set_shrt = set_bcp_shrts;
			if(lemma.size() == 2){
				lemma.push_back(FALSE_LIT); asrt(ufix<=2);
				set_shrt = true;
			}else{
				asrt(confl_lbd >=1); asrt(lemma.size()>2);
				if(ufix>2){glob.registrLBD(confl_lbd);}
			}

			asrt(lemma.size()>2);
			ClsP learnt_cl = clsCreate(lemma,ufix,true,confl_lbd,set_shrt);
			enqueue(lemma[0], learnt_cl,confl_lbd);
        }
    }

    //maximal number of conflicts reached
    return UNKNOWN_TrBool;

}




uint32_t BasicSolver::backtrackLevel(){
    asrt(ass_info.litAssLev(lemma[0]) == decisionLevel());

    uint32_t lsz 	= lemma.size();
    if(lsz < 2){ 	return 0;}

    uint32_t max_idx = 1;
    while(--lsz){
    	uint32_t lev = ass_info.litAssLev(lemma[lsz]);
    	zeroLevel(lev); var_flags.carelessClear(lemma[lsz].toVar());
        if(lev > ass_info.litAssLev(lemma[max_idx])){max_idx = lsz;}
    }

    //such that the literals with maximum level are the watchers
    //this is important for backtracking!
    Literal highest		= lemma[max_idx];
    lemma[max_idx]     	= lemma[1];
    lemma[1]           	= highest;

    asrt(var_flags.allCleared()); sanity(confl_levs.allCleared());

    return ass_info.litAssLev(lemma[1]);
}


Literal BasicSolver::nextDecision(const bool rand){

	Variable var = INVALID_VAR;
    if(rand){
		//random decision
		do{
			if(var_act_hp.empty()) return INVALID_LIT;
			var     = var_act_hp.removeAtPos(glob.randy.nextInt(var_act_hp.size()));

		}while(vals.varKnwn(var));

		cnt_stat(rnd_decs,1);

	}else{
		//activity based decision
		do{
			if(var_act_hp.empty()) return INVALID_LIT;
			var     = var_act_hp.removeTop();
		}while(vals.varKnwn(var) );

		cnt_stat(act_decs,1);
	}

	LOG2("Next decision chosen var " << var.toDIMACS());

	asrt(var != INVALID_VAR );

	//uint32_t pol_mode 	= glob.sets.pol_mode;//isExt(var) ? glob.sets.pol_mode_ext : glob.sets.pol_mode;
	bool pol 			= glob.sets.pol_mode  == 2 ? vals.varPrevValWasTrue(var) : glob.sets.pol_mode;
	return  var.toLit(pol);
}


void BasicSolver::resolveZeros(){
	asrt(decisionLevel() != 0); asrt(dec_lev_end[0] > 0);
	setVarOf(FALSE_LIT);
	clear_later.clear();

#ifndef RESOLVE_LEMMA_ZERO
	for(uint32_t i=0; i!=lemma_zro.size();++i){
		Variable var = lemma_zro[i].toVar();
		if(!var_flags.setFlagNew(var)){ continue;}
		const Reason& r = ass_info.varAssignReason(var);
		if(!r.isUnitFix()){
			varBump(var);
			lemma.push_back(lemma_zro[i]);
		}else{
			clear_later.push_back(var);
		}
	}

#else

	while(!lemma_zro.empty()){
		Literal lit	= lemma_zro.back(); lemma_zro.pop_back(); asrt(vals.litFalse(lit));
		Variable vr	= lit.toVar();
		if(var_flags.isSet(vr)){continue;}
		var_flags.setFlag(vr);

		const Reason& r = ass_info.varAssignReason(vr); asrt(r.getLevel()==0); asrt(!r.isDecision());
		if(r.isClsReason()){
			Cls& cls	= glob.getCls(r.getClsReason());asrt(vals.litFalse(cls.otherWatcher(lit.flipped())));
			lemma_zro	.push_back(cls.otherWatcher(lit.flipped()));
			FOR_UW_LITS_STRT(cls,li);
				lemma_zro.push_back(li); asrt(vals.litFalse(li));
			FOR_UW_LITS_END;
		}else if(!r.isUnitFix()){
			varBump(vr); asrt(vals.litFalse(lit));
			lemma.push_back(lit);
			continue;
		}
		clear_later.push_back(vr);
	}
#endif

	clearLaterNow();
	var_flags.carelessClear(FALSE_LIT.toVar());
}


void BasicSolver::firstUIP(int32_t mrkd_cnt){
	asrt(mrkd_cnt > 1); LOG2("Start firstUIP with number of marked lits " << mrkd_cnt);
	glob.stats.cnt_confl++;

	T_ON(conflana);
    const uint32_t   dlev    = decisionLevel(); asrt(dlev > 0);
    uint32_t        ass_idx  = assign_stck.size() ;
    Literal 	pivot_lit;

    while(true){

    	//skip all unseen assignments
    	//since there always exists a UIP the index cannot become negative
    	while (!var_flags.isSet(assign_stck[--ass_idx].toVar())){}
    	asrt(ass_idx >= 0);

    	pivot_lit                       = assign_stck[ass_idx]; asrt(vals.litTrue(pivot_lit));
    	//can be already cleared since this lit cannot appear in any remaining (previous) reason because literals
    	//in reason clauses are all false (except for the one it is the assign reason)
    	var_flags                       .clear(pivot_lit.toVar());

    	if(--mrkd_cnt == 0){
    		LOG2("Leave firstUIP lemma size " << lemma.size() << " and lemma_zro " << lemma_zro.size() << " asserting is " << pivot_lit.flipped().toDIMACS());
    		asrt(ass_info.litAssLev(pivot_lit) == dlev);asrt(pivot_lit != INVALID_LIT);
    		lemma[0]                = pivot_lit.flipped();
    		T_OFF(conflana);
    		return;
    	}

    	LOG2("Check reason for " << pivot_lit.toDIMACS());
    	const Reason& r_info		= ass_info.varAssignReason(pivot_lit.toVar());
    									asrt(!r_info.isDecision()); asrt(r_info.isClsReason());

    	ClsP clp	= r_info.getClsReason();
    	Cls& cl		= glob.getCls(clp); asrt(ass_info.varAssLev(cl.otherWatcher(pivot_lit).toVar())==dlev);

		//do for all except the pivot_lit
    	mrkd_cnt 	+= firstUIPMrkLit(cl.otherWatcher(pivot_lit),dlev);
    	FOR_UW_LITS_STRT(cl,lit);
			mrkd_cnt += firstUIPMrkLit(lit,dlev);
    	FOR_UW_LITS_END;
    	LOG2("Marked reason for " << pivot_lit.toDIMACS());
	}

}


void     BasicSolver::reduceSelfImpl(){

    if(lemma.size() < 2){asrt(var_flags.allCleared()); return;}

    clear_later.clear();
    uint cp,idx;
    for(cp = idx = 1; idx != lemma.size(); ++idx){
        if(litIsNecessary(lemma[idx])) {
        	lemma[cp++] = lemma[idx];}
    }

    lemma.resize(cp);

    clearLaterNow();
}


bool      BasicSolver::litIsNecessary(Literal lit){

    if(ass_info.varIsDecision(lit.toVar())) return true;

    const Variable start_var(lit.toVar()); asrt(reduce_lemma_stck.empty());

    uint32_t cl_end = clear_later.size();
    uint32_t lz_end	= lemma_zro.size();
    bool is_nec;

    while(true){

        asrt(!ass_info.varIsDecision(lit.toVar()));
        const Reason& r_info = ass_info.varAssignReason(lit.toVar()); asrt(r_info.isClsReason() );

		ClsP clp	= r_info.getClsReason();
		Cls& cl		= glob.getCls(clp);

		is_nec 		= reduceSelfImpCheckLiteral(cl.otherWatcher(lit));

		FOR_UW_LITS_STRT(cl,li);
			if(is_nec) break;
			is_nec = reduceSelfImpCheckLiteral(li);
		FOR_UW_LITS_END;

        if(is_nec){
            for( ;cl_end < clear_later.size();++cl_end)   			{var_flags.clear(clear_later[cl_end]);}
            for(int i = reduce_lemma_stck.size();i--;)              {var_flags.clear(reduce_lemma_stck[i].toVar());}
            reduce_lemma_stck	.clear();
            clear_later			.resize(cl_end);
            lemma_zro			.resize(lz_end);
            return true;
        }

        if(reduce_lemma_stck.empty()){
        	clear_later.push_back(start_var);
            return false;
        }

        lit = reduce_lemma_stck.back(); reduce_lemma_stck.pop_back();
        clear_later.push_back(lit.toVar());
    }
}


void BasicSolver::doSimplify(const bool rearrange_clses,const bool bury_sat){

	asrt(decisionLevel() == 0); asrt(glob.isValid());
    LOG2("Do simpl: Props are " << glob.stats.cnt_props << " rearrange is " << rearrange_clses << " bury sat is " << bury_sat);

    T_ON(simpl);
   	opt_istat(simpl_sml,1);

   	cleanDB(max32,max32,rearrange_clses,bury_sat);
   	simpl_next			= glob.stats.cnt_props + glob.stats.litsCount();
   	simplify_assign_idx = assign_stck.size();

   	T_OFF(simpl);
   	//sanity(watching.asrtWtchingOK());
}

/*bool 	BasicSolver::propUnits(vector<Literal>& units){
	asrt(decisionLevel()==0);
	while(!units.empty()){
		Literal lit = units.back(); units.pop_back();
		if(vals.litTrue(lit)) continue;
		if(vals.litFalse(lit)) return false;
		LOG2("Enqueue unit " << lit.toDIMACS() << " after simpl");;
		decide(lit);
	}
	return bcp() == BCP_SUCCESS;
}*/

void		BasicSolver::setPureLits(BitVector<Literal>& seen){
	asrt(decisionLevel()==0);

	for(uint32_t i=0;i!=seen.size();++i){
		Literal lit(i); Variable var = lit.toVar();
		if(seen.isSet(lit) || vals.varKnwn(var) || ass_info.varIsUnit(lit.toVar())) continue;

		LOG1("Pure lit " << lit.toDIMACS());
		asrt(asrtNotContained(lit));
		setUnit(lit.flipped(),CORE_DONT_CARE,true); asrt(ass_info.litFixTrue(lit.flipped()));
		opt_istat(pure_lits,1);
	}
}


uint32_t	BasicSolver::smpleChckUwSat(Cls& cls){
	uint32_t sz = 0;
	FOR_UW_LITS_STRT(cls,lit);
		if(ass_info.litFixTrue(lit)){return max32;} //clause is true
		++sz;
	FOR_UW_LITS_END;
	return sz;
}


uint32_t	BasicSolver::clsRearrange(Cls& cls,WtchEle& wt, const Literal& ow, BitVector<Literal>* seen){
	asrt(seen != NULL);

	lemma_zro	.clear();
	ClsLit *clit = cls.uwBeg();
	ClsLit *cp	 = clit;

	do{
		Literal l(clit->lit);	asrt(!ass_info.litFixTrue(l));
		clit->prop_next = 1; //remove all early breaks for BCP

		seen->setFlag(l);

		if(!ass_info.litFix(l)){
			(cp++)->lit = clit->lit;

		}else if(!ass_info.varUnitFix(l.toVar())){
			lemma_zro.push_back(l);
		}

	}while((clit++)->anyNext()); asrt(cp >= cls.uwBeg());

	ClsLit* pend  = cp;

	for(uint32_t i=0;i!=lemma_zro.size();++i){
		(cp++)->setLit(lemma_zro[i]);}

	if(cp == cls.uwBeg()){(cp++)->setLit(FALSE_LIT);} asrt(cp > cls.uwBeg());

	(cp -1)->setEnd();

	if(pend == cls.uwBeg()){	asrt(ass_info.litFixFalse(cls.front()));
		wt.setBcpShrt(ow);		//only fix unwatched literals
		cls .clearLBDPrt();
	}else{
		asrt(pend > cls.uwBeg());
		(pend-1)->setPropEnd();
	}

	asrt(cp > cls.uwBeg());
	return cp - cls.uwBeg();
}



void	BasicSolver::cleanDB( const uint32_t& max_lbd, const uint32_t& lbd_rnd, const bool rearrange_clses,const bool bury_sat){
	asrt(lbd_rnd <= max_lbd);
	FeedbackForm mem_fdbk;
	glob.c_info.registerFeedbackForm(mem_fdbk);
	glob.stats.cnt_bin_cls	= 0;
	glob.stats.cnt_lcls		= 0;
	glob.stats.cnt_lts		= 0;
	//const bool rand_del		= max_lbd == Reason::LBD_MAX;
	glob.clearLBDDistri();

	//block reasons of current assignments
	for(uint32_t i=1;i!=assign_stck.size();++i){
		const Reason& r	= ass_info.varAssignReason(assign_stck[i].toVar());
		if(!r.isClsReason()){continue;}

		ClsP clp	= r.getClsReason();
		mem_fdbk.markBorder(clp);
	}

	const uint32_t clits = watching.litsCount();
	BitVector<Literal>* seen = NULL;
	if(rearrange_clses){
		LOG1("Big simplification where pure literals are set and prop shortcuts are rearranged");
		seen = new BitVector<Literal>();
		seen->ensureSize(clits,false);
	} //for all literals


	WtchEle wt; uint32_t lbd; Literal ow;
	for(uint32_t l=0;l!=clits;++l){
		Literal lit(l);
		const bool fixt		= ass_info.litFixTrue(lit);
		WIter<WtchEle> wi	= watching.getWtchOfW(lit);

		while(wi.getNext(wt)){

			ClsP clp			= wt.getClsP();
			Cls& cls			= glob.getCls(clp);
			const Literal ow	= cls.otherWatcher(lit);

			//marked for deletion
			if(ow==lit){		goto SKIP_CLS;}

			if(rearrange_clses){	wt.clearBcpShrtIf();}

		    if(lit<ow){ //see it the first time - check sat and remove false literals

		    	//marked from outside since cls is reason for curr assignment
		    	const bool blkd = mem_fdbk.isBorder(clp);

		    	uint32_t uw_cnt = blkd ?
		    						cls.countUnwatched() :
		    						((fixt || ass_info.litFixTrue(ow)) ? max32 : smpleChckUwSat(cls));

		    	//mark for deletion and don't keep in this list
		    	if(uw_cnt == max32){ asrt(!blkd);
		    		if(bury_sat){glob.buryCls(lit,cls,wt.isLrnt() ? getLBD(cls,wt) :  0);}
		    		cls.unsetWatchers();			goto SKIP_CLS;}

		    	if(rearrange_clses){ asrt(seen != NULL);
		    		seen->setFlag(lit);
		    		seen->setFlag(ow);
		    		uw_cnt = clsRearrange(cls,wt,ow,seen);
		    	}

		    	asrt(!rearrange_clses || !wt.isBcpShrt() || ass_info.litFixFalse(cls.front()));
		    	asrt(!rearrange_clses || wt.isBcpShrt() || !ass_info.litFixFalse(cls.front()));

		    	if(wt.isLrnt()){
		    		lbd = getLBD(cls,wt);
		    		if(!blkd && lbd > lbd_rnd && (lbd > max_lbd || glob.randy.nextBool())){
		    			glob.buryCls(lit,cls,lbd);
		    			cls.unsetWatchers();		goto SKIP_CLS;

		    		}//mark for deletion and don't keep in this list
		    		glob.registrLBD(lbd);
		    	}

		    	asrt(uw_cnt > 0);
				mem_fdbk.markBorder(clp); //start of clause
				mem_fdbk.markBorder(clp+uw_cnt); //last word of clause


				cnt_stat(lts,(uw_cnt+2));
			    if(wt.isBcpShrt()){cnt_stat(bin_cls,1);}else{cnt_stat(lcls,1);}

		    }else if(rearrange_clses && ass_info.litFixFalse(cls.front())){
		    	wt.setBcpShrt(ow);
		    }

		    wi.keep(WtchEle(wt,ow));
		    asrt(!rearrange_clses || !wt.isBcpShrt() || ass_info.litFixFalse(cls.front()));
		    asrt(!rearrange_clses || wt.isBcpShrt() || !ass_info.litFixFalse(cls.front()));

		    SKIP_CLS:;
		}
		watching.finish(lit,wi);
	}

	if(seen != NULL){ asrt(rearrange_clses);
		setPureLits(*seen);
		delete seen;
	}

	glob.c_info.submitFeedback(mem_fdbk);
}



void BasicSolver::learntsGC(){
	LOG2("Start Learnts GC");

	T_ON(gc);
	cnt_stat(gc,1);
	uint32_t left	= 0;
	uint32_t right	= Reason::LBD_MAX;
	uint32_t left_c	= glob.lbd_distri[left];

#ifdef GC_LBD_MEDIAN
	uint32_t right_c	= glob.lbd_distri[right];
	while(left<right){
		if(left_c<=right_c){
			left_c 	+= glob.lbd_distri[++left];}
		else{
			right_c += glob.lbd_distri[--right];}
	}
	asrt(left==right);
#else
	uint32_t mx = glob.stats.cnt_confl / 2;
	while(left<right){
		if(left_c >= mx) break;
		left_c 	+= glob.lbd_distri[++left];
	}
#endif
	left -= (left == Reason::LBD_MAX);
	int32_t bury_lbd_max = glob.cemetery.lbd_max;

	if(glob.stats.litsCount() >= glob.sets.max_lits){
		glob.cemetery.lbd_max = 2;
		LOG1("Clean DB reset lbd forced ");
		left = 2;
	}

	//if(left > max_lbd_gc){left = max_lbd_gc++;}
	uint32_t lbd_rnd = left > Reason::LBD_RND ? Reason::LBD_RND : left;


	LOG1("Clean DB at level " << decisionLevel() << " with max_lbd "<<left);
	cleanDB(left , lbd_rnd, false, true);
	glob.cemetery.lbd_max = bury_lbd_max;

	glob.sets.max_learnts	+= glob.sets.max_learnts_inc;
	T_OFF(gc);
}




bool BasicSolver::asrtVerifyModel(const char* file) {
	IFStream in(file);
	if(in.isBad()){
		LOG0("Could not read file to validate Model");
		return false;
	}

	vector<Literal> cls;
	int32_t  parsed_lit; bool ret = true;
	Literal lit = FALSE_LIT; //uint32_t c = 0; uint64_t l = 0; uint32_t lts = 0;

	while(in.hasNext()){
		if(in.currEquals('c')||in.currEquals('p')) {in.skipLine();}
		else{
			bool sat = false;
			while((parsed_lit = in.getNextInt()) != 0){
				lit 	= Literal::DIMACS2Literal(parsed_lit); asrt(parsed_lit == lit.toDIMACS());
				sat 	|= vals.litTrue(lit);
				cls	.push_back(lit);
			}
			if(!sat){
				LOG0("c Model does not sat cls: ");
				for(uint32_t i=0;i!=cls.size();++i){LOG0(cls[i].toDIMACS());}
				ret = false;
			}
			cls.clear();
		}
	}
	LOG0("Verification of model is " << (ret?"supi dupi":"bad"));
	return ret;;
}

bool BasicSolver::asrtClsUnit(const Literal& u, const ClsP& clp){
	if(vals.litKnwn(u)){
		LOG0("Clause " << clp.idx() << " not unit with " << u.toDIMACS());
		return false;
	}

	//cout << " lit unknown " << u.toDIMACS();

	Cls& cl = glob.getCls(clp);
	if(!vals.litFalse(cl.otherWatcher(u))){
		LOG0("Clause " << clp.idx() << " not unit since second watcher not false");
		return false;
	}

	//cout << " lit false " << cl.otherWatcher(u).toDIMACS();

	FOR_UW_LITS_STRT(cl,lit);

		if(!vals.litFalse(lit)){
			LOG0("Clause " << clp.idx() << " not unit since unwatched lit " << lit.toDIMACS() << " not false");
			return false;
		}
		//cout << " lit false " << lit.toDIMACS();
	FOR_UW_LITS_END;

	//cout << endl;

	return true;
}

bool	BasicSolver::asrtOnlyInitialUnitsEnqueued(){

	for(uint32_t i=0;i!=assign_stck.size();++i){
		Literal lit 	= assign_stck[i];
		const Reason& r	= ass_info.varAssignReason(lit.toVar());
		if(!r.isUnit()){
			LOG0("Lit " << lit.toDIMACS() << " enqueued with reason " << (r.isDecision() ? " dec " : " cls"));
			return false;
		}

	}

	return true;
}

bool	BasicSolver::asrtClsDontContain(const Literal& wtch, const ClsP& clp, const Literal& cand){

	if(wtch==cand){
		LOG0("Cls " << clp.idx() << " contains lit " << cand.toDIMACS());return false;}

	Cls& cls = glob.getCls(clp);
	if(cls.otherWatcher(wtch)==cand){
		LOG0("Cls " << clp.idx() << " contains lit " << cand.toDIMACS());return false;}

	FOR_UW_LITS_STRT(cls,lit);
		if(lit==cand){
			LOG0("Cls " << clp.idx() << " contains lit " << cand.toDIMACS());return false;}
	FOR_UW_LITS_END;

	return true;
}

bool	BasicSolver::asrtNotContained(const Literal& cand){
	for(uint32_t l=0;l!=watching.litsCount();++l){
		Literal lit(l);

		WtchEle wt;
		Iter<WtchEle> wi	= watching.getWtchOf(lit);
		while(wi.getNext(wt)){
			ClsP clp		= wt.getClsP();
			if(!asrtClsDontContain(lit,clp,cand)) return false;
		}
	}
	LOG2("No clause contains lit " << cand.toDIMACS());
	return true;
}

bool	BasicSolver::asrtAllFalse(const Literal& wtch, const ClsP& clp, const uint32_t max_lev){

	if(!vals.litFalse(wtch) || ass_info.litAssLev(wtch)>max_lev){
		LOG0("Cls " << clp.idx() << " contains lit " << wtch.toDIMACS() << " that is not false");return false;}

	Cls& cls = glob.getCls(clp);

	Literal ow = cls.otherWatcher(wtch);
	if(!vals.litFalse(ow) ||  ass_info.litAssLev(ow)>max_lev){
		LOG0("Cls " << clp.idx() << " contains lit " << ow.toDIMACS() << " that is not false");return false;}

	FOR_UW_LITS_STRT(cls,lit);
		if(!vals.litFalse(lit)||  ass_info.litAssLev(lit)>max_lev){
			LOG0("Cls " << clp.idx() << " contains lit " << lit.toDIMACS() << " that is not false");return false;}
	FOR_UW_LITS_END;

	return true;
}

bool	BasicSolver::asrtVarEnqued(const Variable& var)const{
	for(uint32_t i=0;i!= assign_stck.size();++i){
		if(assign_stck[i].toVar() == var) return true;
	}
	LOG0("Variable " << var.toDIMACS() << " not in assign stack");
	return false;
}

bool	BasicSolver::asrtVarInCoreCands(const Variable& var)const{
	for(uint32_t i=0;i!= core_cands.size();++i){
		if(core_cands[i].lit.toVar() == var) return true;
	}
	LOG0("Variable " << var.toDIMACS() << " not in core cands");
	return false;
}


bool	BasicSolver::asrtUnitsEnquedOrInCoreCands(){
	bool ok =true;
	for(uint32_t v=0;v!=vals.count();++v){
		Variable var(v);
		if(vals.varKnwn(var)){
			ok = asrtVarEnqued(var) && ok;
		}else if(ass_info.varIsUnit(var)){
			ok = asrtVarInCoreCands(var) && ok;
		}
	}
	return ok;
}

#define val(x) (vals.litTrue(x)?"[true]":(vals.litFalse(x)?"[false]":"[ukwn]"))
bool	BasicSolver::asrtPrintAllClses() const{
	WtchEle wt;

	for(uint32_t l=0;l!=watching.litsCount();++l){
		Literal lit(l);

		Iter<WtchEle> wi	= watching.getWtchOf(lit);
		while(wi.getNext(wt)){
			ClsP clp	= wt.getClsP();
			Cls& cls	= glob.getCls(clp);

			Literal ol  = cls.otherWatcher(lit);
			if(lit > ol) continue;

			cout << lit.toDIMACS() << val(lit) << " ";
			cout << ol.toDIMACS() << val(ol) << " ";
			FOR_UW_LITS_STRT(cls,li);
				cout << li.toDIMACS() << val(li) << " ";
			FOR_UW_LITS_END;
			cout << endl;
		}
	}
	return true;
}


bool	BasicSolver::asrtAllClsesSAT()const{
	WtchEle wt; bool sat = false;

	for(uint32_t l=0;l!=watching.litsCount();++l){
		Literal lit(l);
		if(vals.litTrue(lit)) continue;

		Iter<WtchEle> wi	= watching.getWtchOf(lit);
		while(wi.getNext(wt)){
			sat = false;
			
			ClsP clp	= wt.getClsP();
			Cls& cls	= glob.getCls(clp);
			if(vals.litTrue(cls.otherWatcher(lit))){ continue;}
			FOR_UW_LITS_STRT(cls,li);
				if(vals.litTrue(li)){
					sat = true; break;}
			FOR_UW_LITS_END;
			if(!sat){
				LOG0("Clause not sat by model "); return false;}
		}
	}
	//LOG0("All clauses sat ");
	return true;
}

