/*
 * Cemetery.hpp
 *
 *  Created on: 18.03.2011
 *      Author: stephan
 */

#ifndef CEMETERY_HPP_
#define CEMETERY_HPP_
#include <vector>
#include <list>
#include "Types.hpp"
#include "AssignInfos.hpp"

using std::list; using std::vector;

struct CemeLit{
	uint32_t last	: 1;
	uint32_t lbd	: 1;
	uint32_t lit	: 30;
	CemeLit(const Literal& l, bool LBD):last(0),lbd(LBD),lit(l.idx()){}
	CemeLit(){}
};

class Cemetery{
	static const uint32_t MAX_BLCK 	= MEGA / 4 - 1;
	static const uint32_t ALLC_FRST	= MAX_BLCK / 16;

	list<vector<CemeLit> > blcks ;
	uint32_t next_idx;

	inline void newGrave();
	inline void ensureAdd(const uint32_t& sz);
	inline void bury(const Literal& lit, Cls& cls,const int32_t lbd, const uint32_t max_sz);
public:
	int32_t lbd_max;
	Cemetery():next_idx(0),lbd_max(6){};

	inline void bury(const Literal& a, const Literal& b){
		ensureAdd(2); asrt(!blcks.empty());

		vector<CemeLit>& grave 	= blcks.back();
		grave.push_back(CemeLit(a,0));
		grave.push_back(CemeLit(b,0));
	}
	
	inline void bury(const vector<Literal>& lits);

	inline void bury(const Literal& lit, Cls& cls, const int32_t lbd){
		if(lbd > lbd_max) return;
		bury(lit,cls,lbd,MAX_BLCK);}

	inline bool getNext(vector<Literal>& lts_out,uint32_t& lbd_out, const VarVal& vals);


};

void Cemetery::bury(const vector<Literal>& lits){
	ensureAdd(lits.size()); asrt(!blcks.empty());

	vector<CemeLit>& grave 	= blcks.back();

	for(uint32_t i=0;i!=lits.size();++i){
		grave.push_back(CemeLit(lits[i],0));
	}
	grave.back().last = 1;
}




void Cemetery::bury(const Literal& lit, Cls& cls,const int32_t LBD, const uint32_t max_sz){

	ensureAdd(5); asrt(!blcks.empty());

	vector<CemeLit>& grave 	= blcks.back();
	const uint32_t gsz		= grave.size(); asrt(gsz +2 < max_sz);
	int32_t lbd				= LBD;

	grave.push_back(CemeLit(lit,lbd-- > 0));
	grave.push_back(CemeLit(cls.otherWatcher(lit),lbd-- > 0));

	FOR_UW_LITS_STRT(cls,l);
		if(grave.size() >= max_sz){ //we do not always know the size of the input clause
			grave.resize(gsz);
			newGrave();
			bury(lit,cls,LBD,max32);
			return;
		}
		grave.push_back(CemeLit(l,lbd-- > 0));
	FOR_UW_LITS_END;
	grave.back().last = 1;
}

bool Cemetery::getNext(vector<Literal>& lts_out, uint32_t& lbd_out, const VarVal& vals){

	while(!blcks.empty()){
		vector<CemeLit>& grave 	= blcks.back();
		if(next_idx >= grave.size()){
			grave	.clear();
			next_idx = 0;
			blcks	.pop_back();
			continue;
		}

		lts_out.clear(); CemeLit cl; bool sat = false;
		lbd_out = 0;
		do{
			cl 		= grave[next_idx++];
			lbd_out += cl.lbd;
			Literal lit	(cl.lit);
			if(sat || vals.litTrue(lit)){sat = true; continue;}
			if(!vals.litFalse(lit)) {lts_out.push_back(lit);}
		}while(cl.last == 0);

		if(!sat) return true;
	}
	next_idx=0;
	return false;
}

void Cemetery::ensureAdd(const uint32_t& sz){

	if(blcks.empty() || blcks.back().size() + sz >= MAX_BLCK){
		newGrave();}
}

void Cemetery::newGrave(){
	blcks.push_back(vector<CemeLit>());
	blcks.back().reserve(ALLC_FRST);
	//cout <<  "c New grave, now we have " << blcks.size() << endl;
}




#endif /* CEMETERY_HPP_ */
