/*
 * Queue.hpp
 *
 *  Created on: Jan 15, 2010
 *      Author: kottlers
 */

#ifndef QUEUE_HPP_
#define QUEUE_HPP_
#define queue_hpp_version "$Revision: 1.1 $"

#include "Assertion.hpp"


template<class T,uint32_t InitSize>
class Queue{
	T*			data;
	uint32_t	d_beg;
	uint32_t	d_end;
	uint32_t	m_end;

	//no real copy and assignment
	Queue(const Queue& v):data(v.data),d_beg(v.d_beg),d_end(v.d_end),m_end(v.m_end){}
	Queue& operator=(const Queue& v){return *this;}

	uint32_t		modulo(const uint32_t& p){asrt(p<=m_end);return p*(p<m_end);}
	inline void 	ensureMem();
	inline void		allocMin();
	inline T*		copyElements(T* from, const T* to, T* dest){
					asrt(from<=to); while(from != to){*dest++ = *from++;}return dest;}

public:
	Queue()			:d_beg(0),d_end(0),m_end(0){allocMin();}
	~Queue()		{if(data != NULL){delete[] data;}}

	bool		empty() const			{return d_end==d_beg;}
	T& 	 		front() const			{asrt(!empty()); return data[d_beg];}
	uint32_t	size() const			{return d_beg<=d_end?d_end-d_beg:m_end+d_end-d_beg;}
	inline void push_back(const T& t)	{
				ensureMem();data[d_end]=t;d_end=modulo(d_end+1);asrt(!empty());}
	inline bool pop_front(T& out);
	inline void	clear();


	//if Pred(element) is true the element remains in the queue
	template<class Pred>
	inline	void		filter(Pred& p);

};

template<class T,uint32_t InitSize>
bool Queue<T,InitSize>::pop_front(T& t){
	if(d_beg!=d_end){
		t 		= data[d_beg];
		d_beg 	= modulo(d_beg+1);
		return true;
	}
	clear();
	return false;
}

template<class T,uint32_t InitSize>
void  Queue<T,InitSize>::clear(){
	if(m_end > InitSize){
		delete[] data;
		allocMin();
	}
	d_beg = d_end = 0;
}

template<class T,uint32_t InitSize>
void  Queue<T,InitSize>::allocMin(){
	if(InitSize == 0){data = NULL;return;}

	m_end 	= InitSize;
	data	= new T[m_end];
}

template<class T,uint32_t InitSize>
void  Queue<T,InitSize>::ensureMem(){
	if(modulo(d_end+1) != d_beg) return;

	aux_var(uint32_t sz = size();)
	if(InitSize == 0 && data == NULL){
		asrt(m_end == 0); asrt(InitSize == 0); asrt(d_beg==0);
		m_end 	= 1;
		data	= new T[m_end];
		return;
	}
	asrt(m_end >0);

	uint32_t m	= m_end <<1;
	T* newd		= new T[m]; T* cp;
	if(d_beg <= d_end){
		cp = copyElements(data+d_beg,data+d_end,newd);
	}else{
		cp = copyElements(data+d_beg,data+m_end,newd);
		cp = copyElements(data,data+d_end,cp);
	}
	d_beg		= 0;
	d_end		= cp - newd; asrt(cp >= newd);
	m_end		= m;
	delete[]	data;
	data		= newd;
	asrt(sz==size());
};


template<class T,uint32_t InitSize>
template<class Pred>
void Queue<T,InitSize>::filter(Pred& p){
	const uint32_t end = d_end;
	d_end = d_beg;
	for(uint32_t it=d_beg;it!=end;it=modulo(it+1)){ asrt(it<m_end);
		if(p(data[it])){
			data[d_end]	= data[it];
			d_end		= modulo(d_end+1);
		}
	}
	if(empty()) clear();
}




#endif /* QUEUE_HPP_ */
